#include <msp/gl/extensions/ext_draw_range_elements.h>
#include <msp/gl/extensions/msp_legacy_features.h>
#include <msp/gl/extensions/msp_primitive_restart.h>
#include <msp/gl/extensions/nv_primitive_restart.h>
#include "batch.h"
#include "bindable.h"
#include "buffer.h"
#include "error.h"
#include "mesh.h"
#include "vertexarray.h"

using namespace std;

namespace {

template<typename T>
void append(vector<unsigned char> &data, T i)
{
	data.insert(data.end(), sizeof(T), 0);
	*(T *)(&data[data.size()-sizeof(T)]) = i;
}

template<typename T, typename U>
U convert(T n)
{
	if(!static_cast<T>(~n))
		return ~0;
	else
		return n;
}

template<typename T, typename U>
void expand(vector<unsigned char> &data)
{
	unsigned count = data.size()/sizeof(T);
	data.resize(count*sizeof(U));
	for(unsigned i=count; i--;)
		*(U *)(&data[i*sizeof(U)]) = convert<T, U>(*(T *)(&data[i*sizeof(T)]));
}

template<typename T, typename U>
void shrink(vector<unsigned char> &data)
{
	unsigned count = data.size()/sizeof(T);
	for(unsigned i=0; i<count; ++i)
		*(U *)(&data[i*sizeof(U)]) = convert<T, U>(*(T *)(&data[i*sizeof(T)]));
	data.resize(count*sizeof(U));
}

}

namespace Msp {
namespace GL {

unsigned Batch::restart_index = 0;

Batch::Batch(PrimitiveType t):
	prim_type(t),
	data_type(UNSIGNED_BYTE),
	min_index(0),
	max_index(0),
	restart(false)
{ }

Batch::~Batch()
{
}

void Batch::set_data_type(DataType t)
{
	if(t!=UNSIGNED_BYTE && t!=UNSIGNED_SHORT && t!=UNSIGNED_INT)
		throw invalid_argument("Batch::set_data_type");
	if(t==UNSIGNED_BYTE && max_index>0xFE)
		throw invalid_operation("Batch::set_data_type");
	else if(t==UNSIGNED_SHORT && max_index>0xFFFE)
		throw invalid_operation("Batch::set_data_type");

	if(data_type==UNSIGNED_BYTE && t==UNSIGNED_SHORT)
		expand<unsigned char, unsigned short>(data);
	else if(data_type==UNSIGNED_BYTE && t==UNSIGNED_INT)
		expand<unsigned char, unsigned>(data);
	else if(data_type==UNSIGNED_SHORT && t==UNSIGNED_INT)
		expand<unsigned short, unsigned>(data);
	else if(data_type==UNSIGNED_INT && t==UNSIGNED_BYTE)
		shrink<unsigned, unsigned char>(data);
	else if(data_type==UNSIGNED_INT && t==UNSIGNED_SHORT)
		shrink<unsigned, unsigned short>(data);
	else if(data_type==UNSIGNED_SHORT && t==UNSIGNED_BYTE)
		shrink<unsigned short, unsigned char>(data);

	data_type = t;
	update_offset();
	dirty = true;
}

Batch &Batch::append(unsigned i)
{
	append_index(i);

	update_offset();
	dirty = true;

	return *this;
}

Batch &Batch::append(const vector<unsigned> &ind)
{
	if(ind.empty())
		return *this;

	data.reserve(data.size()+ind.size()*get_index_size());
	for(vector<unsigned>::const_iterator i=ind.begin(); i!=ind.end(); ++i)
		append_index(*i);

	update_offset();
	dirty = true;

	return *this;
}

bool Batch::can_append(PrimitiveType other_type)
{
	if(other_type!=prim_type)
		return false;
	else if(prim_type==LINE_STRIP || prim_type==LINE_LOOP || prim_type==TRIANGLE_FAN)
		return MSP_primitive_restart || NV_primitive_restart;
	else
		return true;
}

Batch &Batch::append(const Batch &other)
{
	if(other.prim_type!=prim_type)
		throw invalid_argument("Batch::append");
	if(prim_type==LINE_STRIP || prim_type==LINE_LOOP || prim_type==TRIANGLE_FAN)
	{
		if(!MSP_primitive_restart)
		{
			static Require _req(NV_primitive_restart);
			// Make sure we have glEnable/DisableClientState as well
			static Require _req2(MSP_legacy_features);
		}
	}

	if(other.data.empty())
		return *this;

	if(prim_type==POINTS || prim_type==LINES || prim_type==TRIANGLES || prim_type==QUADS)
		;
	else if(MSP_primitive_restart || NV_primitive_restart)
	{
		restart = true;
		if(data_type==UNSIGNED_SHORT)
			::append<unsigned short>(data, 0xFFFF);
		else if(data_type==UNSIGNED_INT)
			::append<unsigned>(data, 0xFFFFFFFF);
		else
			data.push_back(0xFF);
	}
	else if(prim_type==TRIANGLE_STRIP)
	{
		append(get_index(size()-1));
		append(other.get_index(0));
		if(size()&1)
			append(other.get_index(0));
	}
	else if(prim_type==QUAD_STRIP)
	{
		append(get_index(size()-1));
		append(get_index(size()-1));
		append(other.get_index(0));
		append(other.get_index(0));
	}

	unsigned count = other.size();
	for(unsigned i=0; i<count; ++i)
		append_index(other.get_index(i));

	update_offset();
	dirty = true;

	return *this;
}

void Batch::append_index(unsigned i)
{
	if(data.empty())
		min_index = max_index = i;
	else
	{
		min_index = min(min_index, i);
		max_index = max(max_index, i);
	}

	if((data_type==UNSIGNED_BYTE || data_type==UNSIGNED_SHORT) && max_index>0xFFFE)
		set_data_type(UNSIGNED_INT);
	else if(data_type==UNSIGNED_BYTE && max_index>0xFE)
		set_data_type(UNSIGNED_SHORT);

	if(data_type==UNSIGNED_SHORT)
		::append<unsigned short>(data, i);
	else if(data_type==UNSIGNED_INT)
		::append<unsigned>(data, i);
	else
		data.push_back(i);
}

unsigned Batch::get_index_size() const
{
	if(data_type==UNSIGNED_SHORT)
		return sizeof(unsigned short);
	else if(data_type==UNSIGNED_INT)
		return sizeof(unsigned);
	return sizeof(unsigned char);
}

unsigned Batch::get_index(unsigned i) const
{
	if(data_type==UNSIGNED_SHORT)
		return *(unsigned short *)&data[i*sizeof(unsigned short)];
	else if(data_type==UNSIGNED_INT)
		return *(unsigned *)&data[i*sizeof(unsigned )];
	else
		return data[i];
}

void Batch::draw() const
{
	if(restart)
	{
		unsigned index;
		if(data_type==UNSIGNED_SHORT)
			index = 0xFFFF;
		else if(data_type==UNSIGNED_INT)
			index = 0xFFFFFFFF;
		else
			index = 0xFF;

		if(index!=restart_index)
			set_restart_index(index);
	}
	else if(restart_index && restart_index<=max_index)
		set_restart_index(0);

	Buffer *ibuf = get_buffer();
	const void *data_ptr;
	BindRestore _bind_ibuf(ibuf, ELEMENT_ARRAY_BUFFER);
	if(ibuf)
	{
		if(dirty)
			update_buffer();

		data_ptr = reinterpret_cast<const void *>(get_offset());
	}
	else
		data_ptr = &data[0];

	if(EXT_draw_range_elements)
		glDrawRangeElements(prim_type, min_index, max_index, size(), data_type, data_ptr);
	else
		glDrawElements(prim_type, size(), data_type, data_ptr);
}

void Batch::set_restart_index(unsigned index)
{
	if(MSP_primitive_restart)
	{
		if(index>0)
		{
			if(!restart_index)
				glEnable(GL_PRIMITIVE_RESTART);
			glPrimitiveRestartIndex(index);
		}
		else
			glDisable(GL_PRIMITIVE_RESTART);
	}
	else
	{
		if(index>0)
		{
			if(!restart_index)
				glEnableClientState(GL_PRIMITIVE_RESTART_NV);
			glPrimitiveRestartIndexNV(index);
		}
		else
			glDisableClientState(GL_PRIMITIVE_RESTART_NV);
	}

	restart_index = index;
}


Batch::Loader::Loader(Batch &b):
	DataFile::ObjectLoader<Batch>(b)
{
	add("indices", &Loader::indices);
}

void Batch::Loader::indices(const vector<unsigned> &ind)
{
	obj.append(ind);
}

} // namespace GL
} // namespace Msp
