#ifndef MSP_GL_OBJECT_H_
#define MSP_GL_OBJECT_H_

#include <vector>
#include "bindable.h"
#include "renderable.h"
#include "renderpass.h"
#include "resourceobserver.h"

namespace Msp {
namespace GL {

class Material;
class Mesh;
class ObjectInstance;
class Technique;
class Texture;

/**
Combines a Mesh with a Technique to give it an appearance.  The Technique will
define which render passes the Object supports.

In many cases, it's desirable to include multiple copies of an Object in a
Scene, with different model matrices.  ObjectInstances can be used to alter the
rendering of an object on a per-instance basis.

Objects can have multiple levels of detail.  The most detailed level has index
0, with increasing indices having less detail.  When rendering an instance, the
instance's get_level_of_detail method is called to determine which LoD to use.
*/
class Object: public Renderable, private ResourceObserver
{
private:
	struct LevelOfDetail;

	class LodLoader: public DataFile::CollectionObjectLoader<Object>
	{
	private:
		unsigned index;
		LevelOfDetail &lod;

	public:
		LodLoader(Object &, unsigned, Collection *);

	private:
		void mesh(const std::string &);
		void mesh_inline();
		void technique(const std::string &);
		void technique_inline();
	};

public:
	class Loader: public LodLoader
	{
	public:
		Loader(Object &);
		Loader(Object &, Collection &);
	private:
		void init();
		virtual void finish();

		void bounding_sphere_hint(float, float, float, float);
		void level_of_detail(unsigned);
	};

private:
	struct LevelOfDetail
	{
		RefPtr<const Mesh> mesh;
		RefPtr<const Technique> technique;
	};

	std::vector<LevelOfDetail> lods;
	Geometry::BoundingSphere<float, 3> bounding_sphere;
	bool lod0_watched;

public:
	Object();
	Object(const Mesh *, const Technique *);
	~Object();

private:
	LevelOfDetail &get_lod(unsigned, const char *);

public:
	/** Sets the mesh for the highest level of detail (index 0). */
	void set_mesh(const Mesh *m) { set_mesh(0, m); }

	/** Sets the mesh for a given level of detail.  Previous LoDs must have been
	defined. */
	void set_mesh(unsigned, const Mesh *);

private:
	void update_bounding_sphere();
public:
	const Mesh *get_mesh(unsigned = 0) const;

	/** Sets the technique for the highest level of detail (index 0). */
	void set_technique(const Technique *t) { set_technique(0, t); }

	/** Sets the technique for a given level of detail.  Previous LoDs must have
	been defined. */
	void set_technique(unsigned, const Technique *);

	const Technique *get_technique(unsigned = 0) const;
	unsigned get_n_lods() const { return lods.size(); }

	virtual const Geometry::BoundingSphere<float, 3> *get_bounding_sphere() const { return &bounding_sphere; }

	virtual void render(Renderer &, const Tag & = Tag()) const;

	/** Renders an instance of the object.  The instance's hook functions are
	called before and after drawing the mesh. */
	virtual void render(Renderer &, const ObjectInstance &, const Tag & = Tag()) const;

protected:
	virtual void setup_render(Renderer &, const Tag &) const { }
	virtual void finish_render(Renderer &, const Tag &) const { }

private:
	const RenderPass *get_pass(const Tag &, unsigned) const;

	virtual void resource_loaded(Resource &);
	virtual void resource_removed(Resource &);
};

} // namespace GL
} // namespace Msp

#endif
